# import stuff
from Globals import DTMLFile, MessageDialog, Persistent, data_dir
from AccessControl import getSecurityManager
from StringIO import StringIO

# some other stuff
import OFS.SimpleItem
import Acquisition
import AccessControl.Role
import string, base64, sys, os, time

# do xml via basic auth, yeah!
import xmlrpclibBasicAuth

# needed for diff
from OFS.History import html_diff, replace
from string import join, split, atoi, strip
from DocumentTemplate.DT_Util import html_quote
from DocumentTemplate.DT_Var import url_quote_plus
from types import StringType
from Config import diffable, colours, error_messages, syncable, reserved_names

manage_addServerForm = DTMLFile('dtml/Add', globals()) 

def manage_addServer(self, id='', title='', REQUEST=None):
    """ adds a server """
    self._setObject(id, ZSyncer(id, title))
    if REQUEST is not None:
        return MessageDialog(title = 'Added', message = "ZSyncer sucessfully added.", action = 'manage_main', )

class ZSyncer(OFS.SimpleItem.Item, Persistent, Acquisition.Implicit, AccessControl.Role.RoleManager):
    meta_type = 'ZSyncer'

    manage_options = (
        {'label': 'Edit Server', 'action': 'manage_edit' },
        {'label': 'Sync', 'action': 'manage_sync' },
        {'label': 'Security', 'action': 'manage_access' },
    ) 

    __ac_permissions__= (
        ('View management screens', (
        'manage_tabs','manage_main','manage_edit','is_diffable', 'manage_approvedAction', 'manage_addXMLRPC',
        'manage_listObjects', 'manage_exportXMLRPC', 'manage_deleteXMLRPC', 'manage_diffObject',
        'manage_compare', 'manage_srcXMLRPC', 'status_colour', 'manage_folders', 'getSyncable') ),
        ('Change permissions', ('manage_access',) ),
    )

    manage_main = manage_edit = DTMLFile('dtml/Edit', globals())
    try: manage_edit._setName("manage_edit")
    except AttributeError: pass # python 1.5.2

    manage_sync = DTMLFile('dtml/Sync', globals())
    manage_approval = DTMLFile('dtml/Approval', globals())
    manage_diff = DTMLFile('dtml/Diff', globals())
    manage_folders = DTMLFile('dtml/Folder', globals())

    def __init__(self, id, title=''):
        """initialize a new instance of Server"""
        self.id = id
        self.title = title
        self.source_server = [ None, ] # source is this Syncer
        self.dest_server = [ ]
        self.log = 0
        self.approval = 0
        self.override = 0
        self.user = ''
        self.pwd = ''
        self.syncable = syncable
        self.logfile = os.path.join(data_dir, 'ZSyncer.log')

    def getSyncable(self):
        """ Get the synclist, mostly here so upgrades from 0.3 work """
        if not hasattr(self, 'syncable'):
            self.syncable = syncable
        return self.syncable

    
    def manage_approvedAction(self, action, object, syncer=None, approver=None, comments=None):
        """ If this action is to be approved, throw that data in into the message list """
        msgs = ['Syncer: %s' % syncer,]
        msgs.append('Approver: %s' % approver)

        # string out any newlines in comments
        msgs.append('Comments: %s' % string.replace(comments, '\n', ''))

        # call the relevant thing
        if action == 'manage_exportXMLRPC': return self.manage_exportXMLRPC(object, msgs=msgs)
        elif action == 'manage_deleteXMLRPC': return self.manage_deleteXMLRPC(object, msgs=msgs)
        else: return self._error(msg='Unknown action')

    def manage_editServer(self, title='', dest=None, log=0, approval=0, override=0, user='', pwd='', logfile='', REQUEST=None, syncable=syncable):
        """ Allows editing of the server configuration """
        self.title = title
        self.log = log
        self.logfile = logfile
        self.approval = approval
        self.syncable = syncable
        self.dest_server = []
        # lets do a bit of input checking
        for d in dest:
            d = string.strip(d)
            if d: self.dest_server.append(d)
        self.override = override
        self.user = user
        self.pwd = pwd
        if REQUEST is not None:
            return MessageDialog(title = 'Edited', message = "Properties changed.", action = 'manage_edit')

    def manage_deleteXMLRPC(self, object, msgs=None, REQUEST=None):
        """ Interface to calling delete on each destination server via XMLRPC"""
        if msgs is None: msgs = []
        if type(object) == StringType: object = [object,]

        for ob in object:
            for srv in self.dest_server:
                # more logging
                msgs.append('Object %s deleting from' % (ob, srv))
                msgs.append(self._deleteXMLRPC(ob, srv))
        return self._msg(msgs, REQUEST)

    def manage_exportXMLRPC(self, object, msgs=None, REQUEST=None):
        """ Interface to calling export on each object to destination server via XMLRPC """
        if msgs is None: msgs = []
        if type(object) == StringType: object = [object,]

        for ob in object:
            for srv in self.dest_server:
                # more logging
                msgs.append('Object %s to %s' % (ob, srv))
                msgs.append(self._exportXMLRPC(ob, srv))
        return self._msg(msgs, REQUEST)

    def manage_diffObject(self, object, REQUEST=None):
        """ Try to get the object from the destination server and diff on it """
        source = self.manage_srcXMLRPC(object)
        dest = self._srcXMLRPC(self.dest_server[0], object)
        
        # hmm check result, should be a better way to do this using raise, oh well
        if type(source) == type(1): return self._msg([source,], REQUEST)
        if type(dest) == type(1): return self._msg([dest,], REQUEST)
        return html_diff(dest, source)

    def manage_compare(self, folder):
        """ compares the zopes by getting a list from each"""
        try:
            source = self.manage_listObjects(folder)
            dest = self._get_list(self.dest_server[0], folder)
        except IndexError:
            raise 'Server Error', 'There was a problem syncing the servers'
        return self._compare_lists(source, dest)

    def manage_srcXMLRPC(self, obj_path):
        """ Get a src from an object (if allowed) suitable for diffing """
        try: obj = self.restrictedTraverse(obj_path)
        except KeyError: return 404

        c = getSecurityManager().checkPermission
        if not c('View management screens', obj):
            return 403

        return obj.document_src()

    def manage_addXMLRPC(self, data, obj_path, add_in=1):
        """ Adds an object into the server """
        # make sure we always get a string or a list
        if type(obj_path) == type('string'): obj_path = string.split(obj_path, '/')

        # object lets try finding the parent
        try: parent = self.restrictedTraverse(obj_path[:-1])
        except KeyError: return 404

        # lets check they are allowed to do this
        c = getSecurityManager().checkPermission
        allowed = 1
        for perm in ['Delete objects', 'Import/Export objects']:
            if not c(perm, parent): allowed = 0
        if not allowed: return 403

        # if there is one there already, delete it
        if obj_path[-1] in parent.objectIds(): 
            parent.manage_delObjects([obj_path[-1],])
        else:
            if not add_in: # ok so if we are deleting
                return 404

        # lets do it
        if add_in:
            # fake a file using StringIO
            file = StringIO()
            file.write(self._decode(data))
            file.seek(0)

            # now import
            new = parent._p_jar.importFile(file)
            parent._setObject(new.getId(), new)

        # yay!
        return 200

    def status_colour(self, status):
        ''' gives a list of status colours for pretty html '''
        return colours.get(status, 'white')

    def manage_listObjects(self, folder):
        ''' Gets a list of the objects '''
        obs = {}
        # go find the folder, then iterate through the subject
        for ob in self.restrictedTraverse(string.split(folder, '/')).objectValues():
            # we dont to sync ZSyncers, that way leads much confusion
            # so we'll just sync objects we know work
            # avoid broken objects
            if not self._isSyncable(ob.meta_type): continue
            if ob.meta_type[:14] == 'Broken Because': continue
            if self._isReserved(ob.getPhysicalPath()): continue

            o = {}
            o['id'] = ob.getId()
            o['path'] = ob.getPhysicalPath()             
            o['title'] = ob.title_or_id()
            o['meta_type'] = ob.meta_type
            o['icon'] = ob.icon
           
            try: o['is_folder'] = ob.isPrincipiaFolderish
            except KeyError: o['is_folder'] = 0
            
            o['last_modified_time'] = str(ob.bobobase_modification_time())
            
            # add this to the list of objects
            obs[o['id']] = o
        return obs

    def is_diffable(self, meta_type=None):
        ''' is this diffable '''
        if meta_type in diffable: return 1
        else: return 0


###############################
# internal only
# protected by having _ as the first character
    def _isSyncable(self, m):
       if m in self.getSyncable():
            return 1

    def _isReserved(self, p):
        if len(p) == 2 and p[1] in reserved_names:
            return 1

    def _get_list(self, server, folder):
        ''' actually get the list '''
        serverconn = self._serverConn(self._check_http(server))
        return serverconn.manage_listObjects(folder)

    def _compare_lists(self, s_dict, d_dict):
        ''' compare the lists from the two zopes '''
        _l = []
        d_keys = d_dict.keys()
        s_keys = s_dict.keys()

        for item in s_dict.keys():
            # ok its not the dest, hence its missing
            if s_dict[item]['id'] not in d_keys: 
                s_dict[item]['status'] = 'missing'
                _l.append(s_dict[item])
            # ok its on the dest, but the last modified time is earlier
            elif s_dict[item]['last_modified_time'] > d_dict[item]['last_modified_time']:
                s_dict[item]['status'] = 'out of date'
                _l.append(s_dict[item])
            # not sure what else to check
            else: 
                s_dict[item]['status'] = 'ok'
                _l.append(s_dict[item])
        # ok check its on the destination, but not on the original
        for item in d_dict.keys():
            if d_dict[item]['id'] not in s_keys: 
                d_dict[item]['status'] = 'extra'
                _l.append(d_dict[item])
        return _l

    def _encode(self, s):
        # do any other encoding here
        # eg twofish, zlib etc
        # this will do me for now
        return base64.encodestring(s)

    def _decode(self, s):
        # do any other encoding here
        return base64.decodestring(s)

    def _serverConn(self, server):
        # connect using xmlrpc lib
        u = self._get_user()
        return xmlrpclibBasicAuth.Server(server, u['user'], u['pwd'])

    def _error(self, code=None, msg=None):
        # oops
        if msg is None and code is not None: msg = error_messages.get(code, 'Unknown error occured')
        return MessageDialog(title='Message', message=msg, action='manage_main')

    def _get_user(self):
        # get user information
        if self.override:
            _u = {'user':self.user, 'pwd':self.pwd }
        else:
            u = getSecurityManager().getUser()
            _u = {'user':u.getId(), 'pwd':u._getPassword()}
        return _u

    def _deleteXMLRPC(self, object=None, server_url=None, server_port=80, dest_syncer=None):
        """exports an object"""
        # now transfer it
        serverconn = self._serverConn(self._check_http(server_url))
        try: result = serverconn.manage_addXMLRPC('', object, 0)
        except: return 500
        return result

    def _check_http(self, v):
        # so people can get away with leaving http:// off :)
        if v[:7] != 'http://': v = 'http://%s' % v
        if v[:5] != 'http:' and v[:6] != 'https:': v = 'http:%s' % v
        return v

    def _get_time(self):
        # get time for logging
        # could be done using DateTime, but i think I want to fiddle this
        return time.asctime(time.localtime(time.time()))

    def _log(self, msgs):
        # log to ZSyncer
        # this will become more configurable
        file = open(self.logfile, 'a')

        # write out the log
        file.write('%s\tSyncing started\n' % self._get_time())
        for m in msgs: file.write('%s\t%s\n' % (self._get_time(), m))
        file.write('%s\tSyncing finished\n' % self._get_time())

        file.close()

    def _msg(self, msgs, REQUEST=None):
        # what to do with these messages
        ms = []
        for m in msgs: 
            # hack-o-rama, if i have a number, its a status code, so lets look it up and get a nicer message
            if type(m) == type(1): ms.append('%s: %s' % (m, error_messages.get(str(m), 'Unknown error occured')))
            elif m[:6] == 'Object': ms.append('<li>%s' % m)
            else: ms.append(m)

        if self.log or self.approval: self._log(ms)
        if REQUEST is not None:
            return REQUEST.RESPONSE.redirect(
                    'manage_sync?folder=%s&msg=%s' % 
                    (
                        self.REQUEST.get('folder', '/'), 
                        url_quote_plus(join(ms, '<br>'))
                    )
                )

    def _srcXMLRPC(self, server, object=None):
        # call the src method
        serverconn = self._serverConn(self._check_http(server))
        return serverconn.manage_srcXMLRPC(object)

    def _exportXMLRPC(self, object=None, dest_url=None, add_in=1):
        """exports an object"""
        # go get the object
        obj = self.restrictedTraverse(object)
        data = StringIO()
        obj._p_jar.exportFile(obj._p_oid, data)

        # do xml-rpc
        serverconn = self._serverConn(self._check_http(dest_url))
        try: result = serverconn.manage_addXMLRPC(self._encode(data.getvalue()), obj.getPhysicalPath(), add_in)
        except: return 500

        # return result
        return result
